/* Unicode Script IDs */
/* generated from https://www.unicode.org/Public/17.0.0/ucd/UCD.zip:Scripts.txt */
/* DO NOT EDIT!! */

enum unicode_script_id {
    Script_Unknown = 0,
    Script_Common = 1,
    Script_Inherited = 2,
    Script_Arabic = 3,
    Script_Armenian = 4,
    Script_Avestan = 5,
    Script_Balinese = 6,
    Script_Bamum = 7,
    Script_Batak = 8,
    Script_Bengali = 9,
    Script_Bopomofo = 10,
    Script_Brahmi = 11,
    Script_Braille = 12,
    Script_Buginese = 13,
    Script_Buhid = 14,
    Script_Canadian_Aboriginal = 15,
    Script_Carian = 16,
    Script_Cham = 17,
    Script_Cherokee = 18,
    Script_Coptic = 19,
    Script_Cuneiform = 20,
    Script_Cypriot = 21,
    Script_Cyrillic = 22,
    Script_Deseret = 23,
    Script_Devanagari = 24,
    Script_Egyptian_Hieroglyphs = 25,
    Script_Ethiopic = 26,
    Script_Georgian = 27,
    Script_Glagolitic = 28,
    Script_Gothic = 29,
    Script_Greek = 30,
    Script_Gujarati = 31,
    Script_Gurmukhi = 32,
    Script_Han = 33,
    Script_Hangul = 34,
    Script_Hanunoo = 35,
    Script_Hebrew = 36,
    Script_Hiragana = 37,
    Script_Imperial_Aramaic = 38,
    Script_Inscriptional_Pahlavi = 39,
    Script_Inscriptional_Parthian = 40,
    Script_Javanese = 41,
    Script_Kaithi = 42,
    Script_Kannada = 43,
    Script_Katakana = 44,
    Script_Kayah_Li = 45,
    Script_Kharoshthi = 46,
    Script_Khmer = 47,
    Script_Lao = 48,
    Script_Latin = 49,
    Script_Lepcha = 50,
    Script_Limbu = 51,
    Script_Linear_B = 52,
    Script_Lisu = 53,
    Script_Lycian = 54,
    Script_Lydian = 55,
    Script_Malayalam = 56,
    Script_Mandaic = 57,
    Script_Meetei_Mayek = 58,
    Script_Mongolian = 59,
    Script_Myanmar = 60,
    Script_New_Tai_Lue = 61,
    Script_Nko = 62,
    Script_Ogham = 63,
    Script_Ol_Chiki = 64,
    Script_Old_Italic = 65,
    Script_Old_Persian = 66,
    Script_Old_South_Arabian = 67,
    Script_Old_Turkic = 68,
    Script_Oriya = 69,
    Script_Osmanya = 70,
    Script_Phags_Pa = 71,
    Script_Phoenician = 72,
    Script_Rejang = 73,
    Script_Runic = 74,
    Script_Samaritan = 75,
    Script_Saurashtra = 76,
    Script_Shavian = 77,
    Script_Sinhala = 78,
    Script_Sundanese = 79,
    Script_Syloti_Nagri = 80,
    Script_Syriac = 81,
    Script_Tagalog = 82,
    Script_Tagbanwa = 83,
    Script_Tai_Le = 84,
    Script_Tai_Tham = 85,
    Script_Tai_Viet = 86,
    Script_Tamil = 87,
    Script_Telugu = 88,
    Script_Thaana = 89,
    Script_Thai = 90,
    Script_Tibetan = 91,
    Script_Tifinagh = 92,
    Script_Ugaritic = 93,
    Script_Vai = 94,
    Script_Yi = 95,
    Script_Chakma = 96,
    Script_Meroitic_Cursive = 97,
    Script_Meroitic_Hieroglyphs = 98,
    Script_Miao = 99,
    Script_Sharada = 100,
    Script_Sora_Sompeng = 101,
    Script_Takri = 102,
    Script_Bassa_Vah = 103,
    Script_Caucasian_Albanian = 104,
    Script_Duployan = 105,
    Script_Elbasan = 106,
    Script_Grantha = 107,
    Script_Khojki = 108,
    Script_Khudawadi = 109,
    Script_Linear_A = 110,
    Script_Mahajani = 111,
    Script_Manichaean = 112,
    Script_Mende_Kikakui = 113,
    Script_Modi = 114,
    Script_Mro = 115,
    Script_Nabataean = 116,
    Script_Old_North_Arabian = 117,
    Script_Old_Permic = 118,
    Script_Pahawh_Hmong = 119,
    Script_Palmyrene = 120,
    Script_Pau_Cin_Hau = 121,
    Script_Psalter_Pahlavi = 122,
    Script_Siddham = 123,
    Script_Tirhuta = 124,
    Script_Warang_Citi = 125,
    Script_Adlam = 126,
    Script_Ahom = 127,
    Script_Anatolian_Hieroglyphs = 128,
    Script_Bhaiksuki = 129,
    Script_Hatran = 130,
    Script_Marchen = 131,
    Script_Multani = 132,
    Script_Newa = 133,
    Script_Old_Hungarian = 134,
    Script_Osage = 135,
    Script_SignWriting = 136,
    Script_Tangut = 137,
    Script_Masaram_Gondi = 138,
    Script_Nushu = 139,
    Script_Soyombo = 140,
    Script_Zanabazar_Square = 141,
    Script_Dogra = 142,
    Script_Gunjala_Gondi = 143,
    Script_Hanifi_Rohingya = 144,
    Script_Makasar = 145,
    Script_Medefaidrin = 146,
    Script_Old_Sogdian = 147,
    Script_Sogdian = 148,
    Script_Elymaic = 149,
    Script_Nyiakeng_Puachue_Hmong = 150,
    Script_Nandinagari = 151,
    Script_Wancho = 152,
    Script_Chorasmian = 153,
    Script_Dives_Akuru = 154,
    Script_Khitan_Small_Script = 155,
    Script_Yezidi = 156,
    Script_Cypro_Minoan = 157,
    Script_Kawi = 158,
    Script_Nag_Mundari = 159,
    Script_Old_Uyghur = 160,
    Script_Tangsa = 161,
    Script_Toto = 162,
    Script_Vithkuqi = 163,
    Script_Garay = 164,
    Script_Gurung_Khema = 165,
    Script_Kirat_Rai = 166,
    Script_Ol_Onal = 167,
    Script_Sunuwar = 168,
    Script_Todhri = 169,
    Script_Tulu_Tigalari = 170,
    Script_LastId = 170
};
