/*
 * Copyright 2007 Mikolaj Zalewski
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "windef.h"
#include "winuser.h"
#include "commctrl.h"

#define IDS_CANCELLING      16
#define IDS_REMAINING1      17
#define IDS_REMAINING2      18
#define IDS_SECONDS         19
#define IDS_MINUTES         20
#define IDS_HOURS           21

#define IDC_ANIMATION           100
#define IDC_PROGRESS_BAR        102
#define IDC_TEXT_LINE           103

#define IDD_PROGRESS_DLG        100
