/*
 * Copyright 2025 Vibhav Pant
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#ifdef __WIDL__
#pragma makedep header winmd
#pragma winrt ns_prefix
#endif

import "inspectable.idl";
import "windowscontracts.idl";

namespace Wine.Test {
    interface ITest1;

    runtimeclass Test1;

    [
        contract(Windows.Foundation.UniversalApiContract, 1.0),
        exclusiveto(Wine.Test.Test1),
        uuid(deadbeef-dead-beef-dead-deadbeef0001)
    ]
    interface ITest1 : IInspectable
    {
        HRESULT Foo();
    }

    [
        contract(Windows.Foundation.UniversalApiContract, 1.0),
        marshaling_behavior(agile)
    ]
    runtimeclass Test1
    {
        [default] interface Wine.Test.ITest1;
    }
}
