/*
 * Copyright 2022 Rémi Bernon for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep header

#ifdef __WIDL__
#pragma winrt ns_prefix
#endif

import "propidl.idl";
import "inspectable.idl";
import "asyncinfo.idl";
import "windowscontracts.idl";

namespace Wine.Internal {
    /* type-pruning version of AsyncOperationCompletedHandlerImpl<T> */
    delegate HRESULT AsyncOperationCompletedHandlerImpl([in] IInspectable *async, [in] AsyncStatus status);

    [
        uuid(d81ab70d-82e0-481c-983d-401225d98a2c)
    ]
    interface IAsyncInfoImpl : IUnknown
    {
        [propput] HRESULT Completed([in] AsyncOperationCompletedHandlerImpl *handler);
        [propget] HRESULT Completed([out, retval] AsyncOperationCompletedHandlerImpl **handler);
        [propget] HRESULT Result([out, retval] PROPVARIANT *result);
        HRESULT Start();
    }

    typedef HRESULT (*async_operation_callback)( IUnknown *invoker, IUnknown *param, PROPVARIANT *result, BOOL called_async );
}
