/* Automatically generated by tools/make_specfiles */

#define ALL_SYSCALLS32 \
    SYSCALL_ENTRY( 0x0000, NtAccessCheck, 32 ) \
    SYSCALL_ENTRY( 0x0001, NtWorkerFactoryWorkerReady, 4 ) \
    SYSCALL_ENTRY( 0x0002, NtAcceptConnectPort, 24 ) \
    SYSCALL_ENTRY( 0x0003, NtMapUserPhysicalPagesScatter, 0 ) \
    SYSCALL_ENTRY( 0x0004, NtWaitForSingleObject, 12 ) \
    SYSCALL_ENTRY( 0x0005, NtCallbackReturn, 12 ) \
    SYSCALL_ENTRY( 0x0006, NtReadFile, 36 ) \
    SYSCALL_ENTRY( 0x0007, NtDeviceIoControlFile, 40 ) \
    SYSCALL_ENTRY( 0x0008, NtWriteFile, 36 ) \
    SYSCALL_ENTRY( 0x0009, NtRemoveIoCompletion, 20 ) \
    SYSCALL_ENTRY( 0x000a, NtReleaseSemaphore, 12 ) \
    SYSCALL_ENTRY( 0x000b, NtReplyWaitReceivePort, 16 ) \
    SYSCALL_ENTRY( 0x000c, NtReplyPort, 8 ) \
    SYSCALL_ENTRY( 0x000d, NtSetInformationThread, 16 ) \
    SYSCALL_ENTRY( 0x000e, NtSetEvent, 8 ) \
    SYSCALL_ENTRY( 0x000f, NtClose, 4 ) \
    SYSCALL_ENTRY( 0x0010, NtQueryObject, 20 ) \
    SYSCALL_ENTRY( 0x0011, NtQueryInformationFile, 20 ) \
    SYSCALL_ENTRY( 0x0012, NtOpenKey, 12 ) \
    SYSCALL_ENTRY( 0x0013, NtEnumerateValueKey, 24 ) \
    SYSCALL_ENTRY( 0x0014, NtFindAtom, 12 ) \
    SYSCALL_ENTRY( 0x0015, NtQueryDefaultLocale, 8 ) \
    SYSCALL_ENTRY( 0x0016, NtQueryKey, 20 ) \
    SYSCALL_ENTRY( 0x0017, NtQueryValueKey, 24 ) \
    SYSCALL_ENTRY( 0x0018, NtAllocateVirtualMemory, 24 ) \
    SYSCALL_ENTRY_NtQueryInformationProcess( 0x0019, NtQueryInformationProcess, 20 ) \
    SYSCALL_ENTRY( 0x001a, NtWaitForMultipleObjects32, 0 ) \
    SYSCALL_ENTRY( 0x001b, NtWriteFileGather, 36 ) \
    SYSCALL_ENTRY( 0x001c, NtSetInformationProcess, 16 ) \
    SYSCALL_ENTRY( 0x001d, NtCreateKey, 28 ) \
    SYSCALL_ENTRY( 0x001e, NtFreeVirtualMemory, 16 ) \
    SYSCALL_ENTRY( 0x001f, NtImpersonateClientOfPort, 8 ) \
    SYSCALL_ENTRY( 0x0020, NtReleaseMutant, 8 ) \
    SYSCALL_ENTRY( 0x0021, NtQueryInformationToken, 20 ) \
    SYSCALL_ENTRY( 0x0022, NtRequestWaitReplyPort, 12 ) \
    SYSCALL_ENTRY( 0x0023, NtQueryVirtualMemory, 24 ) \
    SYSCALL_ENTRY( 0x0024, NtOpenThreadToken, 16 ) \
    SYSCALL_ENTRY( 0x0025, NtQueryInformationThread, 20 ) \
    SYSCALL_ENTRY( 0x0026, NtOpenProcess, 16 ) \
    SYSCALL_ENTRY( 0x0027, NtSetInformationFile, 20 ) \
    SYSCALL_ENTRY( 0x0028, NtMapViewOfSection, 40 ) \
    SYSCALL_ENTRY( 0x0029, NtAccessCheckAndAuditAlarm, 44 ) \
    SYSCALL_ENTRY( 0x002a, NtUnmapViewOfSection, 8 ) \
    SYSCALL_ENTRY( 0x002b, NtReplyWaitReceivePortEx, 20 ) \
    SYSCALL_ENTRY( 0x002c, NtTerminateProcess, 8 ) \
    SYSCALL_ENTRY( 0x002d, NtSetEventBoostPriority, 4 ) \
    SYSCALL_ENTRY( 0x002e, NtReadFileScatter, 36 ) \
    SYSCALL_ENTRY( 0x002f, NtOpenThreadTokenEx, 20 ) \
    SYSCALL_ENTRY( 0x0030, NtOpenProcessTokenEx, 16 ) \
    SYSCALL_ENTRY( 0x0031, NtQueryPerformanceCounter, 8 ) \
    SYSCALL_ENTRY( 0x0032, NtEnumerateKey, 24 ) \
    SYSCALL_ENTRY( 0x0033, NtOpenFile, 24 ) \
    SYSCALL_ENTRY( 0x0034, NtDelayExecution, 8 ) \
    SYSCALL_ENTRY( 0x0035, NtQueryDirectoryFile, 44 ) \
    SYSCALL_ENTRY( 0x0036, NtQuerySystemInformation, 16 ) \
    SYSCALL_ENTRY( 0x0037, NtOpenSection, 12 ) \
    SYSCALL_ENTRY( 0x0038, NtQueryTimer, 20 ) \
    SYSCALL_ENTRY( 0x0039, NtFsControlFile, 40 ) \
    SYSCALL_ENTRY( 0x003a, NtWriteVirtualMemory, 20 ) \
    SYSCALL_ENTRY( 0x003b, NtCloseObjectAuditAlarm, 12 ) \
    SYSCALL_ENTRY( 0x003c, NtDuplicateObject, 28 ) \
    SYSCALL_ENTRY( 0x003d, NtQueryAttributesFile, 8 ) \
    SYSCALL_ENTRY( 0x003e, NtClearEvent, 4 ) \
    SYSCALL_ENTRY( 0x003f, NtReadVirtualMemory, 20 ) \
    SYSCALL_ENTRY( 0x0040, NtOpenEvent, 12 ) \
    SYSCALL_ENTRY( 0x0041, NtAdjustPrivilegesToken, 24 ) \
    SYSCALL_ENTRY( 0x0042, NtDuplicateToken, 24 ) \
    SYSCALL_ENTRY( 0x0043, NtContinue, 8 ) \
    SYSCALL_ENTRY( 0x0044, NtQueryDefaultUILanguage, 4 ) \
    SYSCALL_ENTRY( 0x0045, NtQueueApcThread, 20 ) \
    SYSCALL_ENTRY( 0x0046, NtYieldExecution, 0 ) \
    SYSCALL_ENTRY( 0x0047, NtAddAtom, 12 ) \
    SYSCALL_ENTRY( 0x0048, NtCreateEvent, 20 ) \
    SYSCALL_ENTRY( 0x0049, NtQueryVolumeInformationFile, 20 ) \
    SYSCALL_ENTRY( 0x004a, NtCreateSection, 28 ) \
    SYSCALL_ENTRY( 0x004b, NtFlushBuffersFile, 8 ) \
    SYSCALL_ENTRY( 0x004c, NtApphelpCacheControl, 0 ) \
    SYSCALL_ENTRY( 0x004d, NtCreateProcessEx, 0 ) \
    SYSCALL_ENTRY( 0x004e, NtCreateThread, 32 ) \
    SYSCALL_ENTRY( 0x004f, NtIsProcessInJob, 8 ) \
    SYSCALL_ENTRY( 0x0050, NtProtectVirtualMemory, 20 ) \
    SYSCALL_ENTRY( 0x0051, NtQuerySection, 20 ) \
    SYSCALL_ENTRY( 0x0052, NtResumeThread, 8 ) \
    SYSCALL_ENTRY( 0x0053, NtTerminateThread, 8 ) \
    SYSCALL_ENTRY( 0x0054, NtReadRequestData, 24 ) \
    SYSCALL_ENTRY( 0x0055, NtCreateFile, 44 ) \
    SYSCALL_ENTRY( 0x0056, NtQueryEvent, 20 ) \
    SYSCALL_ENTRY( 0x0057, NtWriteRequestData, 24 ) \
    SYSCALL_ENTRY( 0x0058, NtOpenDirectoryObject, 12 ) \
    SYSCALL_ENTRY( 0x0059, NtAccessCheckByTypeAndAuditAlarm, 64 ) \
    SYSCALL_ENTRY_NtQuerySystemTime( 0x005a, NtQuerySystemTime, 4 ) \
    SYSCALL_ENTRY( 0x005b, NtWaitForMultipleObjects, 20 ) \
    SYSCALL_ENTRY( 0x005c, NtSetInformationObject, 16 ) \
    SYSCALL_ENTRY( 0x005d, NtCancelIoFile, 8 ) \
    SYSCALL_ENTRY( 0x005e, NtTraceEvent, 0 ) \
    SYSCALL_ENTRY( 0x005f, NtPowerInformation, 20 ) \
    SYSCALL_ENTRY( 0x0060, NtSetValueKey, 24 ) \
    SYSCALL_ENTRY( 0x0061, NtCancelTimer, 8 ) \
    SYSCALL_ENTRY( 0x0062, NtSetTimer, 28 ) \
    SYSCALL_ENTRY( 0x0063, NtAdjustGroupsToken, 24 ) \
    SYSCALL_ENTRY( 0x0064, NtAlertMultipleThreadByThreadId, 16 ) \
    SYSCALL_ENTRY( 0x0065, NtAlertResumeThread, 8 ) \
    SYSCALL_ENTRY( 0x0066, NtAlertThread, 4 ) \
    SYSCALL_ENTRY( 0x0067, NtAlertThreadByThreadId, 4 ) \
    SYSCALL_ENTRY( 0x0068, NtAllocateLocallyUniqueId, 4 ) \
    SYSCALL_ENTRY( 0x0069, NtAllocateReserveObject, 12 ) \
    SYSCALL_ENTRY( 0x006a, NtAllocateUuids, 16 ) \
    SYSCALL_ENTRY( 0x006b, NtAllocateVirtualMemoryEx, 28 ) \
    SYSCALL_ENTRY( 0x006c, NtAreMappedFilesTheSame, 8 ) \
    SYSCALL_ENTRY( 0x006d, NtAssignProcessToJobObject, 8 ) \
    SYSCALL_ENTRY( 0x006e, NtCancelIoFileEx, 12 ) \
    SYSCALL_ENTRY( 0x006f, NtCancelSynchronousIoFile, 12 ) \
    SYSCALL_ENTRY( 0x0070, NtCommitTransaction, 8 ) \
    SYSCALL_ENTRY( 0x0071, NtCompareObjects, 8 ) \
    SYSCALL_ENTRY( 0x0072, NtCompareTokens, 12 ) \
    SYSCALL_ENTRY( 0x0073, NtCompleteConnectPort, 4 ) \
    SYSCALL_ENTRY( 0x0074, NtConnectPort, 32 ) \
    SYSCALL_ENTRY( 0x0075, NtContinueEx, 8 ) \
    SYSCALL_ENTRY( 0x0076, NtConvertBetweenAuxiliaryCounterAndPerformanceCounter, 16 ) \
    SYSCALL_ENTRY( 0x0077, NtCreateDirectoryObject, 12 ) \
    SYSCALL_ENTRY( 0x0078, NtCreateIoCompletion, 16 ) \
    SYSCALL_ENTRY( 0x0079, NtCreateJobObject, 12 ) \
    SYSCALL_ENTRY( 0x007a, NtCreateKeyTransacted, 32 ) \
    SYSCALL_ENTRY( 0x007b, NtCreateKeyedEvent, 16 ) \
    SYSCALL_ENTRY( 0x007c, NtCreateLowBoxToken, 36 ) \
    SYSCALL_ENTRY( 0x007d, NtCreateMailslotFile, 32 ) \
    SYSCALL_ENTRY( 0x007e, NtCreateMutant, 16 ) \
    SYSCALL_ENTRY( 0x007f, NtCreateNamedPipeFile, 56 ) \
    SYSCALL_ENTRY( 0x0080, NtCreatePagingFile, 16 ) \
    SYSCALL_ENTRY( 0x0081, NtCreatePort, 20 ) \
    SYSCALL_ENTRY( 0x0082, NtCreateSectionEx, 36 ) \
    SYSCALL_ENTRY( 0x0083, NtCreateSemaphore, 20 ) \
    SYSCALL_ENTRY( 0x0084, NtCreateSymbolicLinkObject, 16 ) \
    SYSCALL_ENTRY( 0x0085, NtCreateThreadEx, 44 ) \
    SYSCALL_ENTRY( 0x0086, NtCreateTimer, 16 ) \
    SYSCALL_ENTRY( 0x0087, NtCreateToken, 52 ) \
    SYSCALL_ENTRY( 0x0088, NtCreateTransaction, 40 ) \
    SYSCALL_ENTRY( 0x0089, NtCreateUserProcess, 44 ) \
    SYSCALL_ENTRY( 0x008a, NtDebugActiveProcess, 8 ) \
    SYSCALL_ENTRY( 0x008b, NtDebugContinue, 12 ) \
    SYSCALL_ENTRY( 0x008c, NtDeleteAtom, 4 ) \
    SYSCALL_ENTRY( 0x008d, NtDeleteFile, 4 ) \
    SYSCALL_ENTRY( 0x008e, NtDeleteKey, 4 ) \
    SYSCALL_ENTRY( 0x008f, NtDeleteValueKey, 8 ) \
    SYSCALL_ENTRY( 0x0090, NtDisplayString, 4 ) \
    SYSCALL_ENTRY( 0x0091, NtFilterToken, 24 ) \
    SYSCALL_ENTRY( 0x0092, NtFlushBuffersFileEx, 20 ) \
    SYSCALL_ENTRY( 0x0093, NtFlushInstructionCache, 12 ) \
    SYSCALL_ENTRY( 0x0094, NtFlushKey, 4 ) \
    SYSCALL_ENTRY( 0x0095, NtFlushProcessWriteBuffers, 0 ) \
    SYSCALL_ENTRY( 0x0096, NtFlushVirtualMemory, 16 ) \
    SYSCALL_ENTRY( 0x0097, NtGetContextThread, 8 ) \
    SYSCALL_ENTRY( 0x0098, NtGetCurrentProcessorNumber, 0 ) \
    SYSCALL_ENTRY( 0x0099, NtGetNextProcess, 20 ) \
    SYSCALL_ENTRY( 0x009a, NtGetNextThread, 24 ) \
    SYSCALL_ENTRY( 0x009b, NtGetNlsSectionPtr, 20 ) \
    SYSCALL_ENTRY( 0x009c, NtGetWriteWatch, 28 ) \
    SYSCALL_ENTRY( 0x009d, NtImpersonateAnonymousToken, 4 ) \
    SYSCALL_ENTRY( 0x009e, NtInitializeNlsFiles, 12 ) \
    SYSCALL_ENTRY( 0x009f, NtInitiatePowerAction, 16 ) \
    SYSCALL_ENTRY( 0x00a0, NtListenPort, 8 ) \
    SYSCALL_ENTRY( 0x00a1, NtLoadDriver, 4 ) \
    SYSCALL_ENTRY( 0x00a2, NtLoadKey, 8 ) \
    SYSCALL_ENTRY( 0x00a3, NtLoadKey2, 12 ) \
    SYSCALL_ENTRY( 0x00a4, NtLoadKeyEx, 32 ) \
    SYSCALL_ENTRY( 0x00a5, NtLockFile, 40 ) \
    SYSCALL_ENTRY( 0x00a6, NtCreateDebugObject, 16 ) \
    SYSCALL_ENTRY( 0x00a7, NtLockVirtualMemory, 16 ) \
    SYSCALL_ENTRY( 0x00a8, NtMakePermanentObject, 4 ) \
    SYSCALL_ENTRY( 0x00a9, NtMakeTemporaryObject, 4 ) \
    SYSCALL_ENTRY( 0x00aa, NtMapViewOfSectionEx, 36 ) \
    SYSCALL_ENTRY( 0x00ab, NtNotifyChangeDirectoryFile, 36 ) \
    SYSCALL_ENTRY( 0x00ac, NtNotifyChangeKey, 40 ) \
    SYSCALL_ENTRY( 0x00ad, NtNotifyChangeMultipleKeys, 48 ) \
    SYSCALL_ENTRY( 0x00ae, NtOpenIoCompletion, 12 ) \
    SYSCALL_ENTRY( 0x00af, NtOpenJobObject, 12 ) \
    SYSCALL_ENTRY( 0x00b0, NtOpenKeyEx, 16 ) \
    SYSCALL_ENTRY( 0x00b1, NtOpenKeyTransacted, 16 ) \
    SYSCALL_ENTRY( 0x00b2, NtOpenKeyTransactedEx, 20 ) \
    SYSCALL_ENTRY( 0x00b3, NtOpenKeyedEvent, 12 ) \
    SYSCALL_ENTRY( 0x00b4, NtOpenMutant, 12 ) \
    SYSCALL_ENTRY( 0x00b5, NtOpenProcessToken, 12 ) \
    SYSCALL_ENTRY( 0x00b6, NtOpenSemaphore, 12 ) \
    SYSCALL_ENTRY( 0x00b7, NtOpenSymbolicLinkObject, 12 ) \
    SYSCALL_ENTRY( 0x00b8, NtOpenThread, 16 ) \
    SYSCALL_ENTRY( 0x00b9, NtOpenTimer, 12 ) \
    SYSCALL_ENTRY( 0x00ba, NtPrivilegeCheck, 12 ) \
    SYSCALL_ENTRY( 0x00bb, NtPulseEvent, 8 ) \
    SYSCALL_ENTRY( 0x00bc, NtQueryDirectoryObject, 28 ) \
    SYSCALL_ENTRY( 0x00bd, NtQueryEaFile, 36 ) \
    SYSCALL_ENTRY( 0x00be, NtQueryFullAttributesFile, 8 ) \
    SYSCALL_ENTRY( 0x00bf, NtQueryInformationAtom, 20 ) \
    SYSCALL_ENTRY( 0x00c0, NtQueryInformationJobObject, 20 ) \
    SYSCALL_ENTRY( 0x00c1, NtQueryInstallUILanguage, 4 ) \
    SYSCALL_ENTRY( 0x00c2, NtQueryIoCompletion, 20 ) \
    SYSCALL_ENTRY( 0x00c3, NtQueryLicenseValue, 20 ) \
    SYSCALL_ENTRY( 0x00c4, NtQueryMultipleValueKey, 24 ) \
    SYSCALL_ENTRY( 0x00c5, NtQueryMutant, 20 ) \
    SYSCALL_ENTRY( 0x00c6, NtQuerySecurityObject, 20 ) \
    SYSCALL_ENTRY( 0x00c7, NtQuerySemaphore, 20 ) \
    SYSCALL_ENTRY( 0x00c8, NtQuerySymbolicLinkObject, 12 ) \
    SYSCALL_ENTRY( 0x00c9, NtQuerySystemEnvironmentValue, 16 ) \
    SYSCALL_ENTRY( 0x00ca, NtQuerySystemEnvironmentValueEx, 20 ) \
    SYSCALL_ENTRY( 0x00cb, NtQuerySystemInformationEx, 24 ) \
    SYSCALL_ENTRY( 0x00cc, NtQueryTimerResolution, 12 ) \
    SYSCALL_ENTRY( 0x00cd, NtQueueApcThreadEx, 24 ) \
    SYSCALL_ENTRY( 0x00ce, NtQueueApcThreadEx2, 28 ) \
    SYSCALL_ENTRY( 0x00cf, NtRaiseException, 12 ) \
    SYSCALL_ENTRY( 0x00d0, NtRaiseHardError, 24 ) \
    SYSCALL_ENTRY( 0x00d1, NtRegisterThreadTerminatePort, 4 ) \
    SYSCALL_ENTRY( 0x00d2, NtReleaseKeyedEvent, 16 ) \
    SYSCALL_ENTRY( 0x00d3, NtRemoveIoCompletionEx, 24 ) \
    SYSCALL_ENTRY( 0x00d4, NtRemoveProcessDebug, 8 ) \
    SYSCALL_ENTRY( 0x00d5, NtRenameKey, 8 ) \
    SYSCALL_ENTRY( 0x00d6, NtReplaceKey, 12 ) \
    SYSCALL_ENTRY( 0x00d7, NtResetEvent, 8 ) \
    SYSCALL_ENTRY( 0x00d8, NtResetWriteWatch, 12 ) \
    SYSCALL_ENTRY( 0x00d9, NtRestoreKey, 12 ) \
    SYSCALL_ENTRY( 0x00da, NtResumeProcess, 4 ) \
    SYSCALL_ENTRY( 0x00db, NtRollbackTransaction, 8 ) \
    SYSCALL_ENTRY( 0x00dc, NtSaveKey, 8 ) \
    SYSCALL_ENTRY( 0x00dd, NtSecureConnectPort, 36 ) \
    SYSCALL_ENTRY( 0x00de, NtSetContextThread, 8 ) \
    SYSCALL_ENTRY( 0x00df, NtSetDebugFilterState, 12 ) \
    SYSCALL_ENTRY( 0x00e0, NtSetDefaultLocale, 8 ) \
    SYSCALL_ENTRY( 0x00e1, NtSetDefaultUILanguage, 4 ) \
    SYSCALL_ENTRY( 0x00e2, NtSetEaFile, 16 ) \
    SYSCALL_ENTRY( 0x00e3, NtSetInformationDebugObject, 20 ) \
    SYSCALL_ENTRY( 0x00e4, NtSetInformationJobObject, 16 ) \
    SYSCALL_ENTRY( 0x00e5, NtSetInformationKey, 16 ) \
    SYSCALL_ENTRY( 0x00e6, NtSetInformationToken, 16 ) \
    SYSCALL_ENTRY( 0x00e7, NtSetInformationVirtualMemory, 24 ) \
    SYSCALL_ENTRY( 0x00e8, NtSetIntervalProfile, 8 ) \
    SYSCALL_ENTRY( 0x00e9, NtSetIoCompletion, 20 ) \
    SYSCALL_ENTRY( 0x00ea, NtSetIoCompletionEx, 24 ) \
    SYSCALL_ENTRY( 0x00eb, NtSetLdtEntries, 24 ) \
    SYSCALL_ENTRY( 0x00ec, NtSetSecurityObject, 12 ) \
    SYSCALL_ENTRY( 0x00ed, NtSetSystemInformation, 12 ) \
    SYSCALL_ENTRY( 0x00ee, NtSetSystemTime, 8 ) \
    SYSCALL_ENTRY( 0x00ef, NtSetThreadExecutionState, 8 ) \
    SYSCALL_ENTRY( 0x00f0, NtSetTimerResolution, 12 ) \
    SYSCALL_ENTRY( 0x00f1, NtSetVolumeInformationFile, 20 ) \
    SYSCALL_ENTRY( 0x00f2, NtShutdownSystem, 4 ) \
    SYSCALL_ENTRY( 0x00f3, NtSignalAndWaitForSingleObject, 16 ) \
    SYSCALL_ENTRY( 0x00f4, NtSuspendProcess, 4 ) \
    SYSCALL_ENTRY( 0x00f5, NtSuspendThread, 8 ) \
    SYSCALL_ENTRY( 0x00f6, NtSystemDebugControl, 24 ) \
    SYSCALL_ENTRY( 0x00f7, NtTerminateJobObject, 8 ) \
    SYSCALL_ENTRY( 0x00f8, NtTestAlert, 0 ) \
    SYSCALL_ENTRY( 0x00f9, NtTraceControl, 24 ) \
    SYSCALL_ENTRY( 0x00fa, NtUnloadDriver, 4 ) \
    SYSCALL_ENTRY( 0x00fb, NtUnloadKey, 4 ) \
    SYSCALL_ENTRY( 0x00fc, NtUnlockFile, 20 ) \
    SYSCALL_ENTRY( 0x00fd, NtUnlockVirtualMemory, 16 ) \
    SYSCALL_ENTRY( 0x00fe, NtUnmapViewOfSectionEx, 12 ) \
    SYSCALL_ENTRY( 0x00ff, NtWaitForAlertByThreadId, 8 ) \
    SYSCALL_ENTRY( 0x0100, NtWaitForDebugEvent, 16 ) \
    SYSCALL_ENTRY( 0x0101, NtWaitForKeyedEvent, 16 ) \
    SYSCALL_ENTRY( 0x0102, NtWow64AllocateVirtualMemory64, 28 ) \
    SYSCALL_ENTRY( 0x0103, NtWow64GetNativeSystemInformation, 16 ) \
    SYSCALL_ENTRY( 0x0104, NtWow64IsProcessorFeaturePresent, 4 ) \
    SYSCALL_ENTRY( 0x0105, NtWow64QueryInformationProcess64, 20 ) \
    SYSCALL_ENTRY( 0x0106, NtWow64ReadVirtualMemory64, 28 ) \
    SYSCALL_ENTRY( 0x0107, NtWow64WriteVirtualMemory64, 28 )
#ifdef _WIN64
#define ALL_SYSCALLS \
    SYSCALL_ENTRY( 0x0000, NtAccessCheck, 64 ) \
    SYSCALL_ENTRY( 0x0001, NtWorkerFactoryWorkerReady, 8 ) \
    SYSCALL_ENTRY( 0x0002, NtAcceptConnectPort, 48 ) \
    SYSCALL_ENTRY( 0x0003, NtMapUserPhysicalPagesScatter, 0 ) \
    SYSCALL_ENTRY( 0x0004, NtWaitForSingleObject, 24 ) \
    SYSCALL_ENTRY( 0x0005, NtCallbackReturn, 24 ) \
    SYSCALL_ENTRY( 0x0006, NtReadFile, 72 ) \
    SYSCALL_ENTRY( 0x0007, NtDeviceIoControlFile, 80 ) \
    SYSCALL_ENTRY( 0x0008, NtWriteFile, 72 ) \
    SYSCALL_ENTRY( 0x0009, NtRemoveIoCompletion, 40 ) \
    SYSCALL_ENTRY( 0x000a, NtReleaseSemaphore, 24 ) \
    SYSCALL_ENTRY( 0x000b, NtReplyWaitReceivePort, 32 ) \
    SYSCALL_ENTRY( 0x000c, NtReplyPort, 16 ) \
    SYSCALL_ENTRY( 0x000d, NtSetInformationThread, 32 ) \
    SYSCALL_ENTRY( 0x000e, NtSetEvent, 16 ) \
    SYSCALL_ENTRY( 0x000f, NtClose, 8 ) \
    SYSCALL_ENTRY( 0x0010, NtQueryObject, 40 ) \
    SYSCALL_ENTRY( 0x0011, NtQueryInformationFile, 40 ) \
    SYSCALL_ENTRY( 0x0012, NtOpenKey, 24 ) \
    SYSCALL_ENTRY( 0x0013, NtEnumerateValueKey, 48 ) \
    SYSCALL_ENTRY( 0x0014, NtFindAtom, 24 ) \
    SYSCALL_ENTRY( 0x0015, NtQueryDefaultLocale, 16 ) \
    SYSCALL_ENTRY( 0x0016, NtQueryKey, 40 ) \
    SYSCALL_ENTRY( 0x0017, NtQueryValueKey, 48 ) \
    SYSCALL_ENTRY( 0x0018, NtAllocateVirtualMemory, 48 ) \
    SYSCALL_ENTRY_NtQueryInformationProcess( 0x0019, NtQueryInformationProcess, 40 ) \
    SYSCALL_ENTRY( 0x001a, NtWaitForMultipleObjects32, 0 ) \
    SYSCALL_ENTRY( 0x001b, NtWriteFileGather, 72 ) \
    SYSCALL_ENTRY( 0x001c, NtSetInformationProcess, 32 ) \
    SYSCALL_ENTRY( 0x001d, NtCreateKey, 56 ) \
    SYSCALL_ENTRY( 0x001e, NtFreeVirtualMemory, 32 ) \
    SYSCALL_ENTRY( 0x001f, NtImpersonateClientOfPort, 16 ) \
    SYSCALL_ENTRY( 0x0020, NtReleaseMutant, 16 ) \
    SYSCALL_ENTRY( 0x0021, NtQueryInformationToken, 40 ) \
    SYSCALL_ENTRY( 0x0022, NtRequestWaitReplyPort, 24 ) \
    SYSCALL_ENTRY( 0x0023, NtQueryVirtualMemory, 48 ) \
    SYSCALL_ENTRY( 0x0024, NtOpenThreadToken, 32 ) \
    SYSCALL_ENTRY( 0x0025, NtQueryInformationThread, 40 ) \
    SYSCALL_ENTRY( 0x0026, NtOpenProcess, 32 ) \
    SYSCALL_ENTRY( 0x0027, NtSetInformationFile, 40 ) \
    SYSCALL_ENTRY( 0x0028, NtMapViewOfSection, 80 ) \
    SYSCALL_ENTRY( 0x0029, NtAccessCheckAndAuditAlarm, 88 ) \
    SYSCALL_ENTRY( 0x002a, NtUnmapViewOfSection, 16 ) \
    SYSCALL_ENTRY( 0x002b, NtReplyWaitReceivePortEx, 40 ) \
    SYSCALL_ENTRY( 0x002c, NtTerminateProcess, 16 ) \
    SYSCALL_ENTRY( 0x002d, NtSetEventBoostPriority, 8 ) \
    SYSCALL_ENTRY( 0x002e, NtReadFileScatter, 72 ) \
    SYSCALL_ENTRY( 0x002f, NtOpenThreadTokenEx, 40 ) \
    SYSCALL_ENTRY( 0x0030, NtOpenProcessTokenEx, 32 ) \
    SYSCALL_ENTRY( 0x0031, NtQueryPerformanceCounter, 16 ) \
    SYSCALL_ENTRY( 0x0032, NtEnumerateKey, 48 ) \
    SYSCALL_ENTRY( 0x0033, NtOpenFile, 48 ) \
    SYSCALL_ENTRY( 0x0034, NtDelayExecution, 16 ) \
    SYSCALL_ENTRY( 0x0035, NtQueryDirectoryFile, 88 ) \
    SYSCALL_ENTRY( 0x0036, NtQuerySystemInformation, 32 ) \
    SYSCALL_ENTRY( 0x0037, NtOpenSection, 24 ) \
    SYSCALL_ENTRY( 0x0038, NtQueryTimer, 40 ) \
    SYSCALL_ENTRY( 0x0039, NtFsControlFile, 80 ) \
    SYSCALL_ENTRY( 0x003a, NtWriteVirtualMemory, 40 ) \
    SYSCALL_ENTRY( 0x003b, NtCloseObjectAuditAlarm, 24 ) \
    SYSCALL_ENTRY( 0x003c, NtDuplicateObject, 56 ) \
    SYSCALL_ENTRY( 0x003d, NtQueryAttributesFile, 16 ) \
    SYSCALL_ENTRY( 0x003e, NtClearEvent, 8 ) \
    SYSCALL_ENTRY( 0x003f, NtReadVirtualMemory, 40 ) \
    SYSCALL_ENTRY( 0x0040, NtOpenEvent, 24 ) \
    SYSCALL_ENTRY( 0x0041, NtAdjustPrivilegesToken, 48 ) \
    SYSCALL_ENTRY( 0x0042, NtDuplicateToken, 48 ) \
    SYSCALL_ENTRY( 0x0043, NtContinue, 16 ) \
    SYSCALL_ENTRY( 0x0044, NtQueryDefaultUILanguage, 8 ) \
    SYSCALL_ENTRY( 0x0045, NtQueueApcThread, 40 ) \
    SYSCALL_ENTRY( 0x0046, NtYieldExecution, 0 ) \
    SYSCALL_ENTRY( 0x0047, NtAddAtom, 24 ) \
    SYSCALL_ENTRY( 0x0048, NtCreateEvent, 40 ) \
    SYSCALL_ENTRY( 0x0049, NtQueryVolumeInformationFile, 40 ) \
    SYSCALL_ENTRY( 0x004a, NtCreateSection, 56 ) \
    SYSCALL_ENTRY( 0x004b, NtFlushBuffersFile, 16 ) \
    SYSCALL_ENTRY( 0x004c, NtApphelpCacheControl, 0 ) \
    SYSCALL_ENTRY( 0x004d, NtCreateProcessEx, 0 ) \
    SYSCALL_ENTRY( 0x004e, NtCreateThread, 64 ) \
    SYSCALL_ENTRY( 0x004f, NtIsProcessInJob, 16 ) \
    SYSCALL_ENTRY( 0x0050, NtProtectVirtualMemory, 40 ) \
    SYSCALL_ENTRY( 0x0051, NtQuerySection, 40 ) \
    SYSCALL_ENTRY( 0x0052, NtResumeThread, 16 ) \
    SYSCALL_ENTRY( 0x0053, NtTerminateThread, 16 ) \
    SYSCALL_ENTRY( 0x0054, NtReadRequestData, 48 ) \
    SYSCALL_ENTRY( 0x0055, NtCreateFile, 88 ) \
    SYSCALL_ENTRY( 0x0056, NtQueryEvent, 40 ) \
    SYSCALL_ENTRY( 0x0057, NtWriteRequestData, 48 ) \
    SYSCALL_ENTRY( 0x0058, NtOpenDirectoryObject, 24 ) \
    SYSCALL_ENTRY( 0x0059, NtAccessCheckByTypeAndAuditAlarm, 128 ) \
    SYSCALL_ENTRY_NtQuerySystemTime( 0x005a, NtQuerySystemTime, 8 ) \
    SYSCALL_ENTRY( 0x005b, NtWaitForMultipleObjects, 40 ) \
    SYSCALL_ENTRY( 0x005c, NtSetInformationObject, 32 ) \
    SYSCALL_ENTRY( 0x005d, NtCancelIoFile, 16 ) \
    SYSCALL_ENTRY( 0x005e, NtTraceEvent, 0 ) \
    SYSCALL_ENTRY( 0x005f, NtPowerInformation, 40 ) \
    SYSCALL_ENTRY( 0x0060, NtSetValueKey, 48 ) \
    SYSCALL_ENTRY( 0x0061, NtCancelTimer, 16 ) \
    SYSCALL_ENTRY( 0x0062, NtSetTimer, 56 ) \
    SYSCALL_ENTRY( 0x0063, NtAdjustGroupsToken, 48 ) \
    SYSCALL_ENTRY( 0x0064, NtAlertMultipleThreadByThreadId, 32 ) \
    SYSCALL_ENTRY( 0x0065, NtAlertResumeThread, 16 ) \
    SYSCALL_ENTRY( 0x0066, NtAlertThread, 8 ) \
    SYSCALL_ENTRY( 0x0067, NtAlertThreadByThreadId, 8 ) \
    SYSCALL_ENTRY( 0x0068, NtAllocateLocallyUniqueId, 8 ) \
    SYSCALL_ENTRY( 0x0069, NtAllocateReserveObject, 24 ) \
    SYSCALL_ENTRY( 0x006a, NtAllocateUuids, 32 ) \
    SYSCALL_ENTRY( 0x006b, NtAllocateVirtualMemoryEx, 56 ) \
    SYSCALL_ENTRY( 0x006c, NtAreMappedFilesTheSame, 16 ) \
    SYSCALL_ENTRY( 0x006d, NtAssignProcessToJobObject, 16 ) \
    SYSCALL_ENTRY( 0x006e, NtCancelIoFileEx, 24 ) \
    SYSCALL_ENTRY( 0x006f, NtCancelSynchronousIoFile, 24 ) \
    SYSCALL_ENTRY( 0x0070, NtCommitTransaction, 16 ) \
    SYSCALL_ENTRY( 0x0071, NtCompareObjects, 16 ) \
    SYSCALL_ENTRY( 0x0072, NtCompareTokens, 24 ) \
    SYSCALL_ENTRY( 0x0073, NtCompleteConnectPort, 8 ) \
    SYSCALL_ENTRY( 0x0074, NtConnectPort, 64 ) \
    SYSCALL_ENTRY( 0x0075, NtContinueEx, 16 ) \
    SYSCALL_ENTRY( 0x0076, NtConvertBetweenAuxiliaryCounterAndPerformanceCounter, 32 ) \
    SYSCALL_ENTRY( 0x0077, NtCreateDirectoryObject, 24 ) \
    SYSCALL_ENTRY( 0x0078, NtCreateIoCompletion, 32 ) \
    SYSCALL_ENTRY( 0x0079, NtCreateJobObject, 24 ) \
    SYSCALL_ENTRY( 0x007a, NtCreateKeyTransacted, 64 ) \
    SYSCALL_ENTRY( 0x007b, NtCreateKeyedEvent, 32 ) \
    SYSCALL_ENTRY( 0x007c, NtCreateLowBoxToken, 72 ) \
    SYSCALL_ENTRY( 0x007d, NtCreateMailslotFile, 64 ) \
    SYSCALL_ENTRY( 0x007e, NtCreateMutant, 32 ) \
    SYSCALL_ENTRY( 0x007f, NtCreateNamedPipeFile, 112 ) \
    SYSCALL_ENTRY( 0x0080, NtCreatePagingFile, 32 ) \
    SYSCALL_ENTRY( 0x0081, NtCreatePort, 40 ) \
    SYSCALL_ENTRY( 0x0082, NtCreateSectionEx, 72 ) \
    SYSCALL_ENTRY( 0x0083, NtCreateSemaphore, 40 ) \
    SYSCALL_ENTRY( 0x0084, NtCreateSymbolicLinkObject, 32 ) \
    SYSCALL_ENTRY( 0x0085, NtCreateThreadEx, 88 ) \
    SYSCALL_ENTRY( 0x0086, NtCreateTimer, 32 ) \
    SYSCALL_ENTRY( 0x0087, NtCreateToken, 104 ) \
    SYSCALL_ENTRY( 0x0088, NtCreateTransaction, 80 ) \
    SYSCALL_ENTRY( 0x0089, NtCreateUserProcess, 88 ) \
    SYSCALL_ENTRY( 0x008a, NtDebugActiveProcess, 16 ) \
    SYSCALL_ENTRY( 0x008b, NtDebugContinue, 24 ) \
    SYSCALL_ENTRY( 0x008c, NtDeleteAtom, 8 ) \
    SYSCALL_ENTRY( 0x008d, NtDeleteFile, 8 ) \
    SYSCALL_ENTRY( 0x008e, NtDeleteKey, 8 ) \
    SYSCALL_ENTRY( 0x008f, NtDeleteValueKey, 16 ) \
    SYSCALL_ENTRY( 0x0090, NtDisplayString, 8 ) \
    SYSCALL_ENTRY( 0x0091, NtFilterToken, 48 ) \
    SYSCALL_ENTRY( 0x0092, NtFlushBuffersFileEx, 40 ) \
    SYSCALL_ENTRY( 0x0093, NtFlushInstructionCache, 24 ) \
    SYSCALL_ENTRY( 0x0094, NtFlushKey, 8 ) \
    SYSCALL_ENTRY( 0x0095, NtFlushProcessWriteBuffers, 0 ) \
    SYSCALL_ENTRY( 0x0096, NtFlushVirtualMemory, 32 ) \
    SYSCALL_ENTRY( 0x0097, NtGetContextThread, 16 ) \
    SYSCALL_ENTRY( 0x0098, NtGetCurrentProcessorNumber, 0 ) \
    SYSCALL_ENTRY( 0x0099, NtGetNextProcess, 40 ) \
    SYSCALL_ENTRY( 0x009a, NtGetNextThread, 48 ) \
    SYSCALL_ENTRY( 0x009b, NtGetNlsSectionPtr, 40 ) \
    SYSCALL_ENTRY( 0x009c, NtGetWriteWatch, 56 ) \
    SYSCALL_ENTRY( 0x009d, NtImpersonateAnonymousToken, 8 ) \
    SYSCALL_ENTRY( 0x009e, NtInitializeNlsFiles, 24 ) \
    SYSCALL_ENTRY( 0x009f, NtInitiatePowerAction, 32 ) \
    SYSCALL_ENTRY( 0x00a0, NtListenPort, 16 ) \
    SYSCALL_ENTRY( 0x00a1, NtLoadDriver, 8 ) \
    SYSCALL_ENTRY( 0x00a2, NtLoadKey, 16 ) \
    SYSCALL_ENTRY( 0x00a3, NtLoadKey2, 24 ) \
    SYSCALL_ENTRY( 0x00a4, NtLoadKeyEx, 64 ) \
    SYSCALL_ENTRY( 0x00a5, NtLockFile, 80 ) \
    SYSCALL_ENTRY( 0x00a6, NtCreateDebugObject, 32 ) \
    SYSCALL_ENTRY( 0x00a7, NtLockVirtualMemory, 32 ) \
    SYSCALL_ENTRY( 0x00a8, NtMakePermanentObject, 8 ) \
    SYSCALL_ENTRY( 0x00a9, NtMakeTemporaryObject, 8 ) \
    SYSCALL_ENTRY( 0x00aa, NtMapViewOfSectionEx, 72 ) \
    SYSCALL_ENTRY( 0x00ab, NtNotifyChangeDirectoryFile, 72 ) \
    SYSCALL_ENTRY( 0x00ac, NtNotifyChangeKey, 80 ) \
    SYSCALL_ENTRY( 0x00ad, NtNotifyChangeMultipleKeys, 96 ) \
    SYSCALL_ENTRY( 0x00ae, NtOpenIoCompletion, 24 ) \
    SYSCALL_ENTRY( 0x00af, NtOpenJobObject, 24 ) \
    SYSCALL_ENTRY( 0x00b0, NtOpenKeyEx, 32 ) \
    SYSCALL_ENTRY( 0x00b1, NtOpenKeyTransacted, 32 ) \
    SYSCALL_ENTRY( 0x00b2, NtOpenKeyTransactedEx, 40 ) \
    SYSCALL_ENTRY( 0x00b3, NtOpenKeyedEvent, 24 ) \
    SYSCALL_ENTRY( 0x00b4, NtOpenMutant, 24 ) \
    SYSCALL_ENTRY( 0x00b5, NtOpenProcessToken, 24 ) \
    SYSCALL_ENTRY( 0x00b6, NtOpenSemaphore, 24 ) \
    SYSCALL_ENTRY( 0x00b7, NtOpenSymbolicLinkObject, 24 ) \
    SYSCALL_ENTRY( 0x00b8, NtOpenThread, 32 ) \
    SYSCALL_ENTRY( 0x00b9, NtOpenTimer, 24 ) \
    SYSCALL_ENTRY( 0x00ba, NtPrivilegeCheck, 24 ) \
    SYSCALL_ENTRY( 0x00bb, NtPulseEvent, 16 ) \
    SYSCALL_ENTRY( 0x00bc, NtQueryDirectoryObject, 56 ) \
    SYSCALL_ENTRY( 0x00bd, NtQueryEaFile, 72 ) \
    SYSCALL_ENTRY( 0x00be, NtQueryFullAttributesFile, 16 ) \
    SYSCALL_ENTRY( 0x00bf, NtQueryInformationAtom, 40 ) \
    SYSCALL_ENTRY( 0x00c0, NtQueryInformationJobObject, 40 ) \
    SYSCALL_ENTRY( 0x00c1, NtQueryInstallUILanguage, 8 ) \
    SYSCALL_ENTRY( 0x00c2, NtQueryIoCompletion, 40 ) \
    SYSCALL_ENTRY( 0x00c3, NtQueryLicenseValue, 40 ) \
    SYSCALL_ENTRY( 0x00c4, NtQueryMultipleValueKey, 48 ) \
    SYSCALL_ENTRY( 0x00c5, NtQueryMutant, 40 ) \
    SYSCALL_ENTRY( 0x00c6, NtQuerySecurityObject, 40 ) \
    SYSCALL_ENTRY( 0x00c7, NtQuerySemaphore, 40 ) \
    SYSCALL_ENTRY( 0x00c8, NtQuerySymbolicLinkObject, 24 ) \
    SYSCALL_ENTRY( 0x00c9, NtQuerySystemEnvironmentValue, 32 ) \
    SYSCALL_ENTRY( 0x00ca, NtQuerySystemEnvironmentValueEx, 40 ) \
    SYSCALL_ENTRY( 0x00cb, NtQuerySystemInformationEx, 48 ) \
    SYSCALL_ENTRY( 0x00cc, NtQueryTimerResolution, 24 ) \
    SYSCALL_ENTRY( 0x00cd, NtQueueApcThreadEx, 48 ) \
    SYSCALL_ENTRY( 0x00ce, NtQueueApcThreadEx2, 56 ) \
    SYSCALL_ENTRY( 0x00cf, NtRaiseException, 24 ) \
    SYSCALL_ENTRY( 0x00d0, NtRaiseHardError, 48 ) \
    SYSCALL_ENTRY( 0x00d1, NtRegisterThreadTerminatePort, 8 ) \
    SYSCALL_ENTRY( 0x00d2, NtReleaseKeyedEvent, 32 ) \
    SYSCALL_ENTRY( 0x00d3, NtRemoveIoCompletionEx, 48 ) \
    SYSCALL_ENTRY( 0x00d4, NtRemoveProcessDebug, 16 ) \
    SYSCALL_ENTRY( 0x00d5, NtRenameKey, 16 ) \
    SYSCALL_ENTRY( 0x00d6, NtReplaceKey, 24 ) \
    SYSCALL_ENTRY( 0x00d7, NtResetEvent, 16 ) \
    SYSCALL_ENTRY( 0x00d8, NtResetWriteWatch, 24 ) \
    SYSCALL_ENTRY( 0x00d9, NtRestoreKey, 24 ) \
    SYSCALL_ENTRY( 0x00da, NtResumeProcess, 8 ) \
    SYSCALL_ENTRY( 0x00db, NtRollbackTransaction, 16 ) \
    SYSCALL_ENTRY( 0x00dc, NtSaveKey, 16 ) \
    SYSCALL_ENTRY( 0x00dd, NtSecureConnectPort, 72 ) \
    SYSCALL_ENTRY( 0x00de, NtSetContextThread, 16 ) \
    SYSCALL_ENTRY( 0x00df, NtSetDebugFilterState, 24 ) \
    SYSCALL_ENTRY( 0x00e0, NtSetDefaultLocale, 16 ) \
    SYSCALL_ENTRY( 0x00e1, NtSetDefaultUILanguage, 8 ) \
    SYSCALL_ENTRY( 0x00e2, NtSetEaFile, 32 ) \
    SYSCALL_ENTRY( 0x00e3, NtSetInformationDebugObject, 40 ) \
    SYSCALL_ENTRY( 0x00e4, NtSetInformationJobObject, 32 ) \
    SYSCALL_ENTRY( 0x00e5, NtSetInformationKey, 32 ) \
    SYSCALL_ENTRY( 0x00e6, NtSetInformationToken, 32 ) \
    SYSCALL_ENTRY( 0x00e7, NtSetInformationVirtualMemory, 48 ) \
    SYSCALL_ENTRY( 0x00e8, NtSetIntervalProfile, 16 ) \
    SYSCALL_ENTRY( 0x00e9, NtSetIoCompletion, 40 ) \
    SYSCALL_ENTRY( 0x00ea, NtSetIoCompletionEx, 48 ) \
    SYSCALL_ENTRY( 0x00eb, NtSetLdtEntries, 48 ) \
    SYSCALL_ENTRY( 0x00ec, NtSetSecurityObject, 24 ) \
    SYSCALL_ENTRY( 0x00ed, NtSetSystemInformation, 24 ) \
    SYSCALL_ENTRY( 0x00ee, NtSetSystemTime, 16 ) \
    SYSCALL_ENTRY( 0x00ef, NtSetThreadExecutionState, 16 ) \
    SYSCALL_ENTRY( 0x00f0, NtSetTimerResolution, 24 ) \
    SYSCALL_ENTRY( 0x00f1, NtSetVolumeInformationFile, 40 ) \
    SYSCALL_ENTRY( 0x00f2, NtShutdownSystem, 8 ) \
    SYSCALL_ENTRY( 0x00f3, NtSignalAndWaitForSingleObject, 32 ) \
    SYSCALL_ENTRY( 0x00f4, NtSuspendProcess, 8 ) \
    SYSCALL_ENTRY( 0x00f5, NtSuspendThread, 16 ) \
    SYSCALL_ENTRY( 0x00f6, NtSystemDebugControl, 48 ) \
    SYSCALL_ENTRY( 0x00f7, NtTerminateJobObject, 16 ) \
    SYSCALL_ENTRY( 0x00f8, NtTestAlert, 0 ) \
    SYSCALL_ENTRY( 0x00f9, NtTraceControl, 48 ) \
    SYSCALL_ENTRY( 0x00fa, NtUnloadDriver, 8 ) \
    SYSCALL_ENTRY( 0x00fb, NtUnloadKey, 8 ) \
    SYSCALL_ENTRY( 0x00fc, NtUnlockFile, 40 ) \
    SYSCALL_ENTRY( 0x00fd, NtUnlockVirtualMemory, 32 ) \
    SYSCALL_ENTRY( 0x00fe, NtUnmapViewOfSectionEx, 24 ) \
    SYSCALL_ENTRY( 0x00ff, NtWaitForAlertByThreadId, 16 ) \
    SYSCALL_ENTRY( 0x0100, NtWaitForDebugEvent, 32 ) \
    SYSCALL_ENTRY( 0x0101, NtWaitForKeyedEvent, 32 )
#else
#define ALL_SYSCALLS ALL_SYSCALLS32
#endif
#define ALL_SYSCALL_STUBS \
    SYSCALL_STUB( NtApphelpCacheControl ) \
    SYSCALL_STUB( NtCreateProcessEx ) \
    SYSCALL_STUB( NtMapUserPhysicalPagesScatter ) \
    SYSCALL_STUB( NtTraceEvent ) \
    SYSCALL_STUB( NtWaitForMultipleObjects32 )
#define SYSCALL_ENTRY_NtQueryInformationProcess(id,name,args) SYSCALL_ENTRY(id,name,args)
#define SYSCALL_ENTRY_NtQuerySystemTime(id,name,args) SYSCALL_ENTRY(id,name,args)
