/* Automatically generated; DO NOT EDIT!! */

#include "winresrc.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
  39616 "#msgctxt#maximum 31 characters#Afghanistan Standard Time"
  39617 "#msgctxt#maximum 31 characters#Afghanistan Daylight Time"
  39618 "(UTC+04:30) Kabul"
  50192 "#msgctxt#maximum 31 characters#Alaskan Standard Time"
  50193 "#msgctxt#maximum 31 characters#Alaskan Daylight Time"
  50194 "(UTC-09:00) Alaska"
  55200 "#msgctxt#maximum 31 characters#Aleutian Standard Time"
  55201 "#msgctxt#maximum 31 characters#Aleutian Daylight Time"
  55202 "(UTC-10:00) Aleutian Islands"
   4464 "#msgctxt#maximum 31 characters#Altai Standard Time"
   4465 "#msgctxt#maximum 31 characters#Altai Daylight Time"
   4466 "(UTC+07:00) Barnaul, Gorno-Altaysk"
  41968 "#msgctxt#maximum 31 characters#Arab Standard Time"
  41969 "#msgctxt#maximum 31 characters#Arab Daylight Time"
  41970 "(UTC+03:00) Kuwait, Riyadh"
  42704 "#msgctxt#maximum 31 characters#Arabian Standard Time"
  42705 "#msgctxt#maximum 31 characters#Arabian Daylight Time"
  42706 "(UTC+04:00) Abu Dhabi, Muscat"
   8768 "#msgctxt#maximum 31 characters#Arabic Standard Time"
   8769 "#msgctxt#maximum 31 characters#Arabic Daylight Time"
   8770 "(UTC+03:00) Baghdad"
  19584 "#msgctxt#maximum 31 characters#Argentina Standard Time"
  19585 "#msgctxt#maximum 31 characters#Argentina Daylight Time"
  19586 "(UTC-03:00) City of Buenos Aires"
  25104 "#msgctxt#maximum 31 characters#Astrakhan Standard Time"
  25105 "#msgctxt#maximum 31 characters#Astrakhan Daylight Time"
  25106 "(UTC+04:00) Astrakhan, Ulyanovsk"
  34224 "#msgctxt#maximum 31 characters#Atlantic Standard Time"
  34225 "#msgctxt#maximum 31 characters#Atlantic Daylight Time"
  34226 "(UTC-04:00) Atlantic Time (Canada)"
  47024 "#msgctxt#maximum 31 characters#AUS Central Standard Time"
  47025 "#msgctxt#maximum 31 characters#AUS Central Daylight Time"
  47026 "(UTC+09:30) Darwin"
  44672 "#msgctxt#maximum 31 characters#Aus Central W. Standard Time"
  44673 "#msgctxt#maximum 31 characters#Aus Central W. Daylight Time"
  44674 "(UTC+08:45) Eucla"
  30368 "#msgctxt#maximum 31 characters#AUS Eastern Standard Time"
  30369 "#msgctxt#maximum 31 characters#AUS Eastern Daylight Time"
  30370 "(UTC+10:00) Canberra, Melbourne, Sydney"
  11312 "#msgctxt#maximum 31 characters#Azerbaijan Standard Time"
  11313 "#msgctxt#maximum 31 characters#Azerbaijan Daylight Time"
  11314 "(UTC+04:00) Baku"
  18832 "#msgctxt#maximum 31 characters#Azores Standard Time"
  18833 "#msgctxt#maximum 31 characters#Azores Daylight Time"
  18834 "(UTC-01:00) Azores"
  50896 "#msgctxt#maximum 31 characters#Bahia Standard Time"
  50897 "#msgctxt#maximum 31 characters#Bahia Daylight Time"
  50898 "(UTC-03:00) Salvador"
  48720 "#msgctxt#maximum 31 characters#Bangladesh Standard Time"
  48721 "#msgctxt#maximum 31 characters#Bangladesh Daylight Time"
  48722 "(UTC+06:00) Dhaka"
  56576 "#msgctxt#maximum 31 characters#Belarus Standard Time"
  56577 "#msgctxt#maximum 31 characters#Belarus Daylight Time"
  56578 "(UTC+03:00) Minsk"
  51968 "#msgctxt#maximum 31 characters#Bougainville Standard Time"
  51969 "#msgctxt#maximum 31 characters#Bougainville Daylight Time"
  51970 "(UTC+11:00) Bougainville Island"
  25712 "#msgctxt#maximum 31 characters#Canada Central Standard Time"
  25713 "#msgctxt#maximum 31 characters#Canada Central Daylight Time"
  25714 "(UTC-06:00) Saskatchewan"
   2000 "#msgctxt#maximum 31 characters#Cape Verde Standard Time"
   2001 "#msgctxt#maximum 31 characters#Cape Verde Daylight Time"
   2002 "(UTC-01:00) Cabo Verde Is."
  29872 "#msgctxt#maximum 31 characters#Caucasus Standard Time"
  29873 "#msgctxt#maximum 31 characters#Caucasus Daylight Time"
  29874 "(UTC+04:00) Yerevan"
  38928 "#msgctxt#maximum 31 characters#Cen. Australia Standard Time"
  38929 "#msgctxt#maximum 31 characters#Cen. Australia Daylight Time"
  38930 "(UTC+09:30) Adelaide"
  36656 "#msgctxt#maximum 31 characters#Central America Standard Time"
  36657 "#msgctxt#maximum 31 characters#Central America Daylight Time"
  36658 "(UTC-06:00) Central America"
   8176 "#msgctxt#maximum 31 characters#Central Asia Standard Time"
   8177 "#msgctxt#maximum 31 characters#Central Asia Daylight Time"
   8178 "(UTC+06:00) Astana"
  56336 "#msgctxt#maximum 31 characters#Central Brazilian Standard Time"
  56337 "#msgctxt#maximum 31 characters#Central Brazilian Daylight Time"
  56338 "(UTC-04:00) Cuiaba"
   4896 "#msgctxt#maximum 31 characters#Central Europe Standard Time"
   4897 "#msgctxt#maximum 31 characters#Central Europe Daylight Time"
   4898 "(UTC+01:00) Belgrade, Bratislava, Budapest, Ljubljana, Prague"
   2976 "#msgctxt#maximum 31 characters#Central European Standard Time"
   2977 "#msgctxt#maximum 31 characters#Central European Daylight Time"
   2978 "(UTC+01:00) Sarajevo, Skopje, Warsaw, Zagreb"
  64416 "#msgctxt#maximum 31 characters#Central Pacific Standard Time"
  64417 "#msgctxt#maximum 31 characters#Central Pacific Daylight Time"
  64418 "(UTC+11:00) Solomon Is., New Caledonia"
  17456 "#msgctxt#maximum 31 characters#Central Standard Time"
  17457 "#msgctxt#maximum 31 characters#Central Daylight Time"
  17458 "(UTC-06:00) Central Time (US & Canada)"
  32800 "#msgctxt#maximum 31 characters#Central Standard Time (Mexico)"
  32801 "#msgctxt#maximum 31 characters#Central Daylight Time (Mexico)"
  32802 "(UTC-06:00) Guadalajara, Mexico City, Monterrey"
  62304 "#msgctxt#maximum 31 characters#Chatham Islands Standard Time"
  62305 "#msgctxt#maximum 31 characters#Chatham Islands Daylight Time"
  62306 "(UTC+12:45) Chatham Islands"
    160 "#msgctxt#maximum 31 characters#China Standard Time"
    161 "#msgctxt#maximum 31 characters#China Daylight Time"
    162 "(UTC+08:00) Beijing, Chongqing, Hong Kong, Urumqi"
  17136 "#msgctxt#maximum 31 characters#Cuba Standard Time"
  17137 "#msgctxt#maximum 31 characters#Cuba Daylight Time"
  17138 "(UTC-05:00) Havana"
  50384 "#msgctxt#maximum 31 characters#Dateline Standard Time"
  50385 "#msgctxt#maximum 31 characters#Dateline Daylight Time"
  50386 "(UTC-12:00) International Date Line West"
  62720 "#msgctxt#maximum 31 characters#E. Africa Standard Time"
  62721 "#msgctxt#maximum 31 characters#E. Africa Daylight Time"
  62722 "(UTC+03:00) Nairobi"
  65264 "#msgctxt#maximum 31 characters#E. Australia Standard Time"
  65265 "#msgctxt#maximum 31 characters#E. Australia Daylight Time"
  65266 "(UTC+10:00) Brisbane"
  32128 "#msgctxt#maximum 31 characters#E. Europe Standard Time"
  32129 "#msgctxt#maximum 31 characters#E. Europe Daylight Time"
  32130 "(UTC+02:00) Chisinau"
  64096 "#msgctxt#maximum 31 characters#E. South America Standard Time"
  64097 "#msgctxt#maximum 31 characters#E. South America Daylight Time"
  64098 "(UTC-03:00) Brasilia"
  58592 "#msgctxt#maximum 31 characters#Easter Island Standard Time"
  58593 "#msgctxt#maximum 31 characters#Easter Island Daylight Time"
  58594 "(UTC-06:00) Easter Island"
  30800 "#msgctxt#maximum 31 characters#Eastern Standard Time"
  30801 "#msgctxt#maximum 31 characters#Eastern Daylight Time"
  30802 "(UTC-05:00) Eastern Time (US & Canada)"
  59536 "#msgctxt#maximum 31 characters#Eastern Standard Time (Mexico)"
  59537 "#msgctxt#maximum 31 characters#Eastern Daylight Time (Mexico)"
  59538 "(UTC-05:00) Chetumal"
  59424 "#msgctxt#maximum 31 characters#Egypt Standard Time"
  59425 "#msgctxt#maximum 31 characters#Egypt Daylight Time"
  59426 "(UTC+02:00) Cairo"
  45952 "#msgctxt#maximum 31 characters#Ekaterinburg Standard Time"
  45953 "#msgctxt#maximum 31 characters#Ekaterinburg Daylight Time"
  45954 "(UTC+05:00) Ekaterinburg"
  25680 "#msgctxt#maximum 31 characters#Fiji Standard Time"
  25681 "#msgctxt#maximum 31 characters#Fiji Daylight Time"
  25682 "(UTC+12:00) Fiji"
  63264 "#msgctxt#maximum 31 characters#FLE Standard Time"
  63265 "#msgctxt#maximum 31 characters#FLE Daylight Time"
  63266 "(UTC+02:00) Helsinki, Kyiv, Riga, Sofia, Tallinn, Vilnius"
    416 "#msgctxt#maximum 31 characters#Georgian Standard Time"
    417 "#msgctxt#maximum 31 characters#Georgian Daylight Time"
    418 "(UTC+04:00) Tbilisi"
   7408 "#msgctxt#maximum 31 characters#GMT Standard Time"
   7409 "#msgctxt#maximum 31 characters#GMT Daylight Time"
   7410 "(UTC+00:00) Dublin, Edinburgh, Lisbon, London"
  58096 "#msgctxt#maximum 31 characters#Greenland Standard Time"
  58097 "#msgctxt#maximum 31 characters#Greenland Daylight Time"
  58098 "(UTC-03:00) Greenland"
  47168 "#msgctxt#maximum 31 characters#Greenwich Standard Time"
  47169 "#msgctxt#maximum 31 characters#Greenwich Daylight Time"
  47170 "(UTC+00:00) Monrovia, Reykjavik"
  24192 "#msgctxt#maximum 31 characters#GTB Standard Time"
  24193 "#msgctxt#maximum 31 characters#GTB Daylight Time"
  24194 "(UTC+02:00) Athens, Bucharest"
   2896 "#msgctxt#maximum 31 characters#Haiti Standard Time"
   2897 "#msgctxt#maximum 31 characters#Haiti Daylight Time"
   2898 "(UTC-05:00) Haiti"
  53376 "#msgctxt#maximum 31 characters#Hawaiian Standard Time"
  53377 "#msgctxt#maximum 31 characters#Hawaiian Daylight Time"
  53378 "(UTC-10:00) Hawaii"
  22400 "#msgctxt#maximum 31 characters#India Standard Time"
  22401 "#msgctxt#maximum 31 characters#India Daylight Time"
  22402 "(UTC+05:30) Chennai, Kolkata, Mumbai, New Delhi"
   5568 "#msgctxt#maximum 31 characters#Iran Standard Time"
   5569 "#msgctxt#maximum 31 characters#Iran Daylight Time"
   5570 "(UTC+03:30) Tehran"
  47968 "#msgctxt#maximum 31 characters#Israel Standard Time"
  47969 "#msgctxt#maximum 31 characters#Israel Daylight Time"
  47970 "(UTC+02:00) Jerusalem"
  17232 "#msgctxt#maximum 31 characters#Jordan Standard Time"
  17233 "#msgctxt#maximum 31 characters#Jordan Daylight Time"
  17234 "(UTC+02:00) Amman"
  14720 "#msgctxt#maximum 31 characters#Kaliningrad Standard Time"
  14721 "#msgctxt#maximum 31 characters#Kaliningrad Daylight Time"
  14722 "(UTC+02:00) Kaliningrad"
  62048 "#msgctxt#maximum 31 characters#Korea Standard Time"
  62049 "#msgctxt#maximum 31 characters#Korea Daylight Time"
  62050 "(UTC+09:00) Seoul"
  50768 "#msgctxt#maximum 31 characters#Libya Standard Time"
  50769 "#msgctxt#maximum 31 characters#Libya Daylight Time"
  50770 "(UTC+02:00) Tripoli"
  16768 "#msgctxt#maximum 31 characters#Line Islands Standard Time"
  16769 "#msgctxt#maximum 31 characters#Line Islands Daylight Time"
  16770 "(UTC+14:00) Kiritimati Island"
   8656 "#msgctxt#maximum 31 characters#Lord Howe Standard Time"
   8657 "#msgctxt#maximum 31 characters#Lord Howe Daylight Time"
   8658 "(UTC+10:30) Lord Howe Island"
   8832 "#msgctxt#maximum 31 characters#Magadan Standard Time"
   8833 "#msgctxt#maximum 31 characters#Magadan Daylight Time"
   8834 "(UTC+11:00) Magadan"
  13120 "#msgctxt#maximum 31 characters#Magallanes Standard Time"
  13121 "#msgctxt#maximum 31 characters#Magallanes Daylight Time"
  13122 "(UTC-03:00) Punta Arenas"
  20736 "#msgctxt#maximum 31 characters#Marquesas Standard Time"
  20737 "#msgctxt#maximum 31 characters#Marquesas Daylight Time"
  20738 "(UTC-09:30) Marquesas Islands"
  60896 "#msgctxt#maximum 31 characters#Mauritius Standard Time"
  60897 "#msgctxt#maximum 31 characters#Mauritius Daylight Time"
  60898 "(UTC+04:00) Port Louis"
  15968 "#msgctxt#maximum 31 characters#Middle East Standard Time"
  15969 "#msgctxt#maximum 31 characters#Middle East Daylight Time"
  15970 "(UTC+02:00) Beirut"
  27968 "#msgctxt#maximum 31 characters#Montevideo Standard Time"
  27969 "#msgctxt#maximum 31 characters#Montevideo Daylight Time"
  27970 "(UTC-03:00) Montevideo"
   2992 "#msgctxt#maximum 31 characters#Morocco Standard Time"
   2993 "#msgctxt#maximum 31 characters#Morocco Daylight Time"
   2994 "(UTC+01:00) Casablanca"
  34352 "#msgctxt#maximum 31 characters#Mountain Standard Time"
  34353 "#msgctxt#maximum 31 characters#Mountain Daylight Time"
  34354 "(UTC-07:00) Mountain Time (US & Canada)"
   7248 "#msgctxt#maximum 31 characters#Mountain Standard Time (Mexico)"
   7249 "#msgctxt#maximum 31 characters#Mountain Daylight Time (Mexico)"
   7250 "(UTC-07:00) Chihuahua, La Paz, Mazatlan"
  21104 "#msgctxt#maximum 31 characters#Myanmar Standard Time"
  21105 "#msgctxt#maximum 31 characters#Myanmar Daylight Time"
  21106 "(UTC+06:30) Yangon (Rangoon)"
  30704 "#msgctxt#maximum 31 characters#N. Central Asia Standard Time"
  30705 "#msgctxt#maximum 31 characters#N. Central Asia Daylight Time"
  30706 "(UTC+07:00) Novosibirsk"
   6208 "#msgctxt#maximum 31 characters#Namibia Standard Time"
   6209 "#msgctxt#maximum 31 characters#Namibia Daylight Time"
   6210 "(UTC+02:00) Windhoek"
   1216 "#msgctxt#maximum 31 characters#Nepal Standard Time"
   1217 "#msgctxt#maximum 31 characters#Nepal Daylight Time"
   1218 "(UTC+05:45) Kathmandu"
  54896 "#msgctxt#maximum 31 characters#New Zealand Standard Time"
  54897 "#msgctxt#maximum 31 characters#New Zealand Daylight Time"
  54898 "(UTC+12:00) Auckland, Wellington"
   9504 "#msgctxt#maximum 31 characters#Newfoundland Standard Time"
   9505 "#msgctxt#maximum 31 characters#Newfoundland Daylight Time"
   9506 "(UTC-03:30) Newfoundland"
  47440 "#msgctxt#maximum 31 characters#Norfolk Standard Time"
  47441 "#msgctxt#maximum 31 characters#Norfolk Daylight Time"
  47442 "(UTC+11:00) Norfolk Island"
  19056 "#msgctxt#maximum 31 characters#North Asia East Standard Time"
  19057 "#msgctxt#maximum 31 characters#North Asia East Daylight Time"
  19058 "(UTC+08:00) Irkutsk"
    304 "#msgctxt#maximum 31 characters#North Asia Standard Time"
    305 "#msgctxt#maximum 31 characters#North Asia Daylight Time"
    306 "(UTC+07:00) Krasnoyarsk"
  35888 "#msgctxt#maximum 31 characters#North Korea Standard Time"
  35889 "#msgctxt#maximum 31 characters#North Korea Daylight Time"
  35890 "(UTC+09:00) Pyongyang"
  55952 "#msgctxt#maximum 31 characters#Omsk Standard Time"
  55953 "#msgctxt#maximum 31 characters#Omsk Daylight Time"
  55954 "(UTC+06:00) Omsk"
  65008 "#msgctxt#maximum 31 characters#Pacific SA Standard Time"
  65009 "#msgctxt#maximum 31 characters#Pacific SA Daylight Time"
  65010 "(UTC-04:00) Santiago"
  11232 "#msgctxt#maximum 31 characters#Pacific Standard Time"
  11233 "#msgctxt#maximum 31 characters#Pacific Daylight Time"
  11234 "(UTC-08:00) Pacific Time (US & Canada)"
  15584 "#msgctxt#maximum 31 characters#Pacific Standard Time (Mexico)"
  15585 "#msgctxt#maximum 31 characters#Pacific Daylight Time (Mexico)"
  15586 "(UTC-08:00) Baja California"
  28752 "#msgctxt#maximum 31 characters#Pakistan Standard Time"
  28753 "#msgctxt#maximum 31 characters#Pakistan Daylight Time"
  28754 "(UTC+05:00) Islamabad, Karachi"
  50224 "#msgctxt#maximum 31 characters#Paraguay Standard Time"
  50225 "#msgctxt#maximum 31 characters#Paraguay Daylight Time"
  50226 "(UTC-04:00) Asuncion"
  53760 "#msgctxt#maximum 31 characters#Qyzylorda Standard Time"
  53761 "#msgctxt#maximum 31 characters#Qyzylorda Daylight Time"
  53762 "(UTC+05:00) Qyzylorda"
  45104 "#msgctxt#maximum 31 characters#Romance Standard Time"
  45105 "#msgctxt#maximum 31 characters#Romance Daylight Time"
  45106 "(UTC+01:00) Brussels, Copenhagen, Madrid, Paris"
  59264 "#msgctxt#maximum 31 characters#Russia Time Zone 10"
  59265 "#msgctxt#maximum 31 characters#Russia Time Zone 10"
  59266 "(UTC+11:00) Chokurdakh"
  46000 "#msgctxt#maximum 31 characters#Russia Time Zone 11"
  46001 "#msgctxt#maximum 31 characters#Russia Time Zone 11"
  46002 "(UTC+12:00) Anadyr, Petropavlovsk-Kamchatsky"
  17920 "#msgctxt#maximum 31 characters#Russia Time Zone 3"
  17921 "#msgctxt#maximum 31 characters#Russia Time Zone 3"
  17922 "(UTC+04:00) Izhevsk, Samara"
  44560 "#msgctxt#maximum 31 characters#Russian Standard Time"
  44561 "#msgctxt#maximum 31 characters#Russian Daylight Time"
  44562 "(UTC+03:00) Moscow, St. Petersburg"
  40656 "#msgctxt#maximum 31 characters#SA Eastern Standard Time"
  40657 "#msgctxt#maximum 31 characters#SA Eastern Daylight Time"
  40658 "(UTC-03:00) Cayenne, Fortaleza"
  49424 "#msgctxt#maximum 31 characters#SA Pacific Standard Time"
  49425 "#msgctxt#maximum 31 characters#SA Pacific Daylight Time"
  49426 "(UTC-05:00) Bogota, Lima, Quito, Rio Branco"
  57120 "#msgctxt#maximum 31 characters#SA Western Standard Time"
  57121 "#msgctxt#maximum 31 characters#SA Western Daylight Time"
  57122 "(UTC-04:00) Georgetown, La Paz, Manaus, San Juan"
   5584 "#msgctxt#maximum 31 characters#Saint Pierre Standard Time"
   5585 "#msgctxt#maximum 31 characters#Saint Pierre Daylight Time"
   5586 "(UTC-03:00) Saint Pierre and Miquelon"
  35104 "#msgctxt#maximum 31 characters#Sakhalin Standard Time"
  35105 "#msgctxt#maximum 31 characters#Sakhalin Daylight Time"
  35106 "(UTC+11:00) Sakhalin"
  14592 "#msgctxt#maximum 31 characters#Samoa Standard Time"
  14593 "#msgctxt#maximum 31 characters#Samoa Daylight Time"
  14594 "(UTC+13:00) Samoa"
   5984 "#msgctxt#maximum 31 characters#Sao Tome Standard Time"
   5985 "#msgctxt#maximum 31 characters#Sao Tome Daylight Time"
   5986 "(UTC+00:00) Sao Tome"
  33808 "#msgctxt#maximum 31 characters#Saratov Standard Time"
  33809 "#msgctxt#maximum 31 characters#Saratov Daylight Time"
  33810 "(UTC+04:00) Saratov"
  53728 "#msgctxt#maximum 31 characters#SE Asia Standard Time"
  53729 "#msgctxt#maximum 31 characters#SE Asia Daylight Time"
  53730 "(UTC+07:00) Bangkok, Hanoi, Jakarta"
  61840 "#msgctxt#maximum 31 characters#Singapore Standard Time"
  61841 "#msgctxt#maximum 31 characters#Singapore Daylight Time"
  61842 "(UTC+08:00) Kuala Lumpur, Singapore"
  38896 "#msgctxt#maximum 31 characters#South Africa Standard Time"
  38897 "#msgctxt#maximum 31 characters#South Africa Daylight Time"
  38898 "(UTC+02:00) Harare, Pretoria"
   7664 "#msgctxt#maximum 31 characters#South Sudan Standard Time"
   7665 "#msgctxt#maximum 31 characters#South Sudan Daylight Time"
   7666 "(UTC+02:00) Juba"
  39408 "#msgctxt#maximum 31 characters#Sri Lanka Standard Time"
  39409 "#msgctxt#maximum 31 characters#Sri Lanka Daylight Time"
  39410 "(UTC+05:30) Sri Jayawardenepura"
   9792 "#msgctxt#maximum 31 characters#Sudan Standard Time"
   9793 "#msgctxt#maximum 31 characters#Sudan Daylight Time"
   9794 "(UTC+02:00) Khartoum"
  46576 "#msgctxt#maximum 31 characters#Syria Standard Time"
  46577 "#msgctxt#maximum 31 characters#Syria Daylight Time"
  46578 "(UTC+02:00) Damascus"
  26432 "#msgctxt#maximum 31 characters#Taipei Standard Time"
  26433 "#msgctxt#maximum 31 characters#Taipei Daylight Time"
  26434 "(UTC+08:00) Taipei"
  36016 "#msgctxt#maximum 31 characters#Tasmania Standard Time"
  36017 "#msgctxt#maximum 31 characters#Tasmania Daylight Time"
  36018 "(UTC+10:00) Hobart"
  43072 "#msgctxt#maximum 31 characters#Tocantins Standard Time"
  43073 "#msgctxt#maximum 31 characters#Tocantins Daylight Time"
  43074 "(UTC-03:00) Araguaina"
  16240 "#msgctxt#maximum 31 characters#Tokyo Standard Time"
  16241 "#msgctxt#maximum 31 characters#Tokyo Daylight Time"
  16242 "(UTC+09:00) Osaka, Sapporo, Tokyo"
  29136 "#msgctxt#maximum 31 characters#Tomsk Standard Time"
  29137 "#msgctxt#maximum 31 characters#Tomsk Daylight Time"
  29138 "(UTC+07:00) Tomsk"
   7216 "#msgctxt#maximum 31 characters#Tonga Standard Time"
   7217 "#msgctxt#maximum 31 characters#Tonga Daylight Time"
   7218 "(UTC+13:00) Nuku'alofa"
  31472 "#msgctxt#maximum 31 characters#Transbaikal Standard Time"
  31473 "#msgctxt#maximum 31 characters#Transbaikal Daylight Time"
  31474 "(UTC+09:00) Chita"
  24848 "#msgctxt#maximum 31 characters#Turkey Standard Time"
  24849 "#msgctxt#maximum 31 characters#Turkey Daylight Time"
  24850 "(UTC+03:00) Istanbul"
  20560 "#msgctxt#maximum 31 characters#Turks And Caicos Standard Time"
  20561 "#msgctxt#maximum 31 characters#Turks And Caicos Daylight Time"
  20562 "(UTC-05:00) Turks and Caicos"
  47200 "#msgctxt#maximum 31 characters#Ulaanbaatar Standard Time"
  47201 "#msgctxt#maximum 31 characters#Ulaanbaatar Daylight Time"
  47202 "(UTC+08:00) Ulaanbaatar"
  34928 "#msgctxt#maximum 31 characters#US Eastern Standard Time"
  34929 "#msgctxt#maximum 31 characters#US Eastern Daylight Time"
  34930 "(UTC-05:00) Indiana (East)"
  37472 "#msgctxt#maximum 31 characters#US Mountain Standard Time"
  37473 "#msgctxt#maximum 31 characters#US Mountain Daylight Time"
  37474 "(UTC-07:00) Arizona"
  22000 "#msgctxt#maximum 31 characters#Coordinated Universal Time"
  22001 "#msgctxt#maximum 31 characters#Coordinated Universal Time"
  22002 "(UTC) Coordinated Universal Time"
   1200 "#msgctxt#maximum 31 characters#UTC+12"
   1201 "#msgctxt#maximum 31 characters#UTC+12"
   1202 "(UTC+12:00) Coordinated Universal Time+12"
   8784 "#msgctxt#maximum 31 characters#UTC+13"
   8785 "#msgctxt#maximum 31 characters#UTC+13"
   8786 "(UTC+13:00) Coordinated Universal Time+13"
  37328 "#msgctxt#maximum 31 characters#UTC-02"
  37329 "#msgctxt#maximum 31 characters#UTC-02"
  37330 "(UTC-02:00) Coordinated Universal Time-02"
   3344 "#msgctxt#maximum 31 characters#UTC-08"
   3345 "#msgctxt#maximum 31 characters#UTC-08"
   3346 "(UTC-08:00) Coordinated Universal Time-08"
  39152 "#msgctxt#maximum 31 characters#UTC-09"
  39153 "#msgctxt#maximum 31 characters#UTC-09"
  39154 "(UTC-09:00) Coordinated Universal Time-09"
  19408 "#msgctxt#maximum 31 characters#UTC-11"
  19409 "#msgctxt#maximum 31 characters#UTC-11"
  19410 "(UTC-11:00) Coordinated Universal Time-11"
  51472 "#msgctxt#maximum 31 characters#Venezuela Standard Time"
  51473 "#msgctxt#maximum 31 characters#Venezuela Daylight Time"
  51474 "(UTC-04:00) Caracas"
  61376 "#msgctxt#maximum 31 characters#Vladivostok Standard Time"
  61377 "#msgctxt#maximum 31 characters#Vladivostok Daylight Time"
  61378 "(UTC+10:00) Vladivostok"
  18048 "#msgctxt#maximum 31 characters#Volgograd Standard Time"
  18049 "#msgctxt#maximum 31 characters#Volgograd Daylight Time"
  18050 "(UTC+04:00) Volgograd"
  65376 "#msgctxt#maximum 31 characters#W. Australia Standard Time"
  65377 "#msgctxt#maximum 31 characters#W. Australia Daylight Time"
  65378 "(UTC+08:00) Perth"
  64976 "#msgctxt#maximum 31 characters#W. Central Africa Standard Time"
  64977 "#msgctxt#maximum 31 characters#W. Central Africa Daylight Time"
  64978 "(UTC+01:00) West Central Africa"
  27696 "#msgctxt#maximum 31 characters#W. Europe Standard Time"
  27697 "#msgctxt#maximum 31 characters#W. Europe Daylight Time"
  27698 "(UTC+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna"
  53952 "#msgctxt#maximum 31 characters#W. Mongolia Standard Time"
  53953 "#msgctxt#maximum 31 characters#W. Mongolia Daylight Time"
  53954 "(UTC+07:00) Hovd"
  49664 "#msgctxt#maximum 31 characters#West Asia Standard Time"
  49665 "#msgctxt#maximum 31 characters#West Asia Daylight Time"
  49666 "(UTC+05:00) Ashgabat, Tashkent"
  46160 "#msgctxt#maximum 31 characters#West Bank Standard Time"
  46161 "#msgctxt#maximum 31 characters#West Bank Daylight Time"
  46162 "(UTC+02:00) Gaza, Hebron"
  10224 "#msgctxt#maximum 31 characters#West Pacific Standard Time"
  10225 "#msgctxt#maximum 31 characters#West Pacific Daylight Time"
  10226 "(UTC+10:00) Guam, Port Moresby"
  40576 "#msgctxt#maximum 31 characters#Yakutsk Standard Time"
  40577 "#msgctxt#maximum 31 characters#Yakutsk Daylight Time"
  40578 "(UTC+09:00) Yakutsk"
  26288 "#msgctxt#maximum 31 characters#Yukon Standard Time"
  26289 "#msgctxt#maximum 31 characters#Yukon Daylight Time"
  26290 "(UTC-07:00) Yukon"
}
