/*
 * Copyright 2026 Nikolay Sivov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep header

import "unknwn.idl";
import "objidl.idl";
import "oaidl.idl";

struct dtd;

[
    object,
    local,
    uuid(437d6b59-62b8-4adb-a9bb-20d5fdc43516)
]
interface ISAXExtensionHandler : IUnknown
{
    HRESULT xmldecl([in] BSTR version, [in] BSTR encoding,[in] BSTR standalone);
    HRESULT dtd([in] BSTR data, [in] struct dtd *dtd);
}

[
    object,
    local,
    uuid(b16ef36b-72ed-4caa-9e2d-91453600d030)
]
interface ISAXXMLReaderExtension : IUnknown
{
    HRESULT parseUTF16([in] ISequentialStream *stream);
}
